@@ -39,16 +39,13 @@ def export_fusion_samples(
     img_limit=1000,
     export_to="asset",
     output_asset_path=None,
-    export_kwargs=None,
     skip_empty=True,
+    seed = 0
 ):
     """
     """
 
-    export_opts = dict(
-        cloud=ee.batch.Export.table.toCloudStorage, asset=ee.batch.Export.table.toAsset,
-    )
-    export_func = export_opts[export_to]
+    optical_indices = ["mndwi","nwi","lswi","aewish","aewinsh","gwi"]
 
     ds_kwargs = dict(region=region, start_time=start_time, end_time=end_time)
     dsa_kwargs = {**ds_kwargs, **{"apply_band_adjustment": True}}
@@ -56,12 +53,16 @@ def export_fusion_samples(
     lc8 = datasets.Landsat8(**ds_kwargs)
     le7 = datasets.Landsat7(**dsa_kwargs)
     s2 = datasets.Sentinel2(**dsa_kwargs)
-    viirs = datasets.Sentinel2(**ds_kwargs)
 
     s1 = datasets.Sentinel1(**ds_kwargs)
     s1 = s1.add_fusion_features()
 
     optical = lc8.merge(s2).merge(le7)
+    optical = optical.apply_func(geeutils.add_indices,
+        indices= optical_indices
+    )
+
+    optical.collection = optical.collection.select(optical_indices)
 
     ds = optical.join(s1)
 
@@ -71,17 +72,6 @@ def export_fusion_samples(
     output_features = ee.FeatureCollection([])
 
     if stratify_samples:
-        # jrc_img = (
-        #     ee.Image("JRC/GSW1_2/GlobalSurfaceWater")
-        #     .select("occurrence")
-        #     .unmask(0)
-        #     .rename("water")
-        # )
-        # stratification_img = (
-        #     jrc_img.where(jrc_img.lt(10), 0)
-        #     .where(jrc_img.gte(10), 1)
-        #     .where(jrc_img.gte(65), 2)
-        # )
         class_band = "landcover"
         igbp_classes = ee.List(
             [1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17]
@@ -90,9 +80,9 @@ def export_fusion_samples(
 
         stratification_img = (
             ee.ImageCollection("MODIS/006/MCD12Q1")
+            .limit(5, "system:time_start", True)
             .mode()
             .remap(igbp_classes, ipcc_classes)
-            .focal_mode()
             .rename(class_band)
         )
         classes = ipcc_classes.distinct()
@@ -114,11 +104,11 @@ def export_fusion_samples(
                     numPoints=n_samples,
                     classBand=class_band,
                     scale=sample_scale,
-                    seed=i,
+                    seed=seed+i,
                     classValues=classes,
                     classPoints=ee.List.repeat(
                         n_samples, classes.size().subtract(1)
-                    ).add(n_samples * 2),
+                    ).add(n_samples * 4),
                     tileScale=16,
                     geometries=True,
                 )
@@ -128,7 +118,7 @@ def export_fusion_samples(
                     region=sample_region,
                     scale=sample_scale,
                     numPixels=n_samples,
-                    seed=i,
+                    seed=seed+i,
                     tileScale=16,
                     geometries=True,
                 )
@@ -145,225 +135,24 @@ def export_fusion_samples(
         except EEException as e:
             break
 
-    export_info = dict(collection=output_features, assetId=output_asset_path)
-    if export_kwargs is not None:
-        export_info = {**export_info, **export_kwargs}
-        # if "fileNamePrefix" in export_info.keys():
-        #     prefix = export_kwargs["fileNamePrefix"]
-        #     true_prefix = (
-        #         prefix + desc
-        #         if prefix.endswith("/")
-        #         else prefix + f"_{dstr}_{i}"
-        #     )
-        #     export_info["fileNamePrefix"] = true_prefix
 
-    task = export_func(collection=output_features, assetId=output_asset_path)
+    task = ee.batch.Export.table.toAsset(collection=output_features, assetId=output_asset_path)
     task.start()
     logging.info(f"Started task")
 
     return
 
-
-def build_fusion_model(
-    features,
-    label,
-    sample_path=None,
-    sample_asset=None,
-    ee_asset_path=None,
-    output_bucket=None,
-    model_type="random forest",
-    filter_outliers=False,
-    seed=0,
-):
-
-    # framework options = [sklearn, xgboost, lightgbm]
-
-    if sample_path.startswith("gs://"):
-        tables = utils.list_gcs_objs(sample_path.replace("gs://", ""), pattern="*.csv")
-    else:
-        raise NotImplementedError(
-            "Currently only fetching data from Google Cloud Storage is supported"
-        )
-
-    df_list = (pd.read_csv(table) for table in tables)
-    df = pd.concat(df_list, axis=0, ignore_index=True)
-
-    X = df[features]
-    y = df[label]
-
-    feature_names = list(X.columns)
-    n_features = len(feature_names)
-
-    if filter_outliers:
-        logging.info(f"applying filters on feature coluns")
-        α = 0.05
-        z_threshold = 3
-        masks = np.ones(df.shape)
-        for i, feature in enumerate(list(X.columns)):
-            values = X[feature]
-            # ratio of number of unique values to the total number of unique values
-            # probability is less than α thereshold then do not filter outliers
-            is_categorical = 1.0 * np.unique(values).size / values.size < α
-            if is_categorical:
-                logging.info(f"{feature} is categorial, passing filter")
-                continue
-
-            k2, p = stats.normaltest(values)
-            if p < α:
-                logging.info(f"{feature} is gaussian, using z-score filter")
-                z = np.abs(stats.zscore(values))
-                masks[:, i] = z < z_threshold
-
-            else:
-                logging.info(f"{feature} is not gaussian, using iqr filter")
-                q1, q3 = stats.iqr(values)
-                iqr = q3 - q1
-                masks[:, i] = (values > (q1 - 1.5 * iqr)) & (values < (q3 + 1.5 * iqr))
-
-        X = X[masks.any(axis=1)]
-        y = y[masks.any(axis=1)]
-
-    if output_training_report:
-        X_train, X_test, y_train, y_test = model_selection.train_test_split(
-            X, y, train_size=0.80, random_state=seed
-        )
-    else:
-        X_train, y_train = X, y
-
-    scaler = preprocessing.MinMaxScaler().fit(X_train)
-    X_train = scaler.transform(X_train)
-
-    fmin = {f"{feature_names[i]}_min": scaler.data_min_[i] for i in range(n_features)}
-    fmax = {f"{feature_names[i]}_max": scaler.data_max_[i] for i in range(n_features)}
-    scaler_fc = ee.FeatureCollection(
-        ee.Feature(ee.Geometry.Point([0, 0]), {**fmin, **fmax})
-    )
-
-    now = datetime.datetime.now()
-    time_id = now.strftime("%Y%m%d%H%M%s")
-
-    if export_scaler_to_ee:
-
-        desc = f"fusion_model_feature_scaling_{time_id}"
-        task = ee.batch.Export.table.toAsset(
-            collection=scaler_fc, description=desc, assetId=ee_asset_path + desc
-        )
-        task.start()
-
-    logging.info(f"training model...")
-    if framework is "sklearn":
-        from sklearn.ensemble import RandomForestRegressor, ExtraTreesRegressor
-
-        # model = RandomForestRegressor(n_estimators=100, n_jobs=-1, random_state=seed,)
-        model = ExtraTreesRegressor(
-            n_estimators=50, n_jobs=-1, max_depth=30, random_state=0
-        )
-
-    else:
-        raise NotImplementedError(
-            "only fusion modeling using the scikit-learn framework is avaiable now"
-        )
-
-    t1 = datetime.datetime.now()
-    model.fit(X_train, y_train)
-    training_time = datetime.datetime.now() - t1
-
-    if output_training_report:
-        X_test = scaler.transform(X_test)
-        y_pred = model.predict(X_test)
-
-        mae = metrics.mean_absolute_error(y_test, y_pred)
-        me = metrics.max_error(y_test, y_pred)
-        r2 = metrics.r2_score(y_test, y_pred)
-        bias = np.mean((y_test - y_pred))
-        rmse = np.mean(np.sqrt((y_test - y_pred) ** 2))
-
-        fs = gcsfs.GCSFileSystem()
-
-        with fs.open(f"{output_bucket}/training_report_{time_id}.txt", "w") as report:
-            content = [
-                f"Model framework: {framework}",
-                f"Execution time: {now}",
-                f"Model parameters: {{\n {pformat(model.get_params(),indent=8)[1:]}",
-                f"Training information:",
-                f"\trandom seed: {seed}",
-                f"\ttraining time: {training_time}",
-                f"\tn training examples: {X_train.shape[0]}",
-                f"\tn testing examples: {X_test.shape[0]}",
-                f"\tbias: {bias}",
-                f"\trmse: {rmse}",
-                f"\tmae: {mae}",
-                f"\tr2: {r2}",
-                f"\tmax error: {me}",
-                f"Feature scaling:",
-                f"\tfeature minimum: {{\n {pformat(fmin,indent=16)[1:]}",
-                f"\tfeature maximum: {{\n {pformat(fmax,indent=16)[1:]}",
-            ]
-            if export_scaler_to_ee:
-                content.append(f"\tOutput scaling feature collection: {desc}")
-
-            report.write("\n".join(content))
-
-    estimators = model.estimators_
-
-    # fs = gcsfs.GCSFileSystem()
-    # for i, estimator in enumerate(estimators):
-    #     string = ml.sklearn_tree_to_string(estimator, features)
-    #     with fs.open(f"{output_bucket}/estimator_{time_id}_{i:04d}.txt", "w") as f:
-    #         f.write(string)
-
-    tree_properties = {}
-    # args = [(est,features) for est in estimators]
-    with mp.Pool(7) as pool:
-        proc = pool.map_async(
-            partial(ml.sklearn_tree_to_string, feature_names=features), estimators
-        )
-        trees = list(proc.get())
-
-    df = pd.DataFrame(
-        {
-            "lat": np.zeros(len(trees)),
-            "lon": np.zeros(len(trees)),
-            "tree": [i.replace("\n", "#") for i in trees],
-        }
-    )
-
-    bucket_obj = f"gs://{output_bucket}/rf_model_{time_id}.csv"
-    df.to_csv(bucket_obj, index=False)
-
-    os.system(
-        f"earthengine upload table {bucket_obj} --asset_id users/kelmarkert/rf_tree_test --x_column lon --y_column lat"
-    )
-
-    # tree_properties = {f"tree_{i:30d}": trees[i] for i in range(len(trees))}
-
-    # features = [ee.Feature(ee.Geometry.Point([0,0])).set('tree',tree_properties[k]) for k in tree_properties.keys()]
-
-    # ee_tree_obj= ee.FeatureCollection(features)
-
-    # task = ee.batch.Export.table.toAsset(
-    #     collection=ee_tree_obj,
-    #     description="rf_tree_export_test",
-    #     assetId="users/kelmarkert/rf_tree_test"
-    # )
-    # task.start()
-
-    return
-
-
 def _fuse_dataset(
     region,
     start_time,
     end_time,
     fusion_model,
     scaling_dict=None,
+    scaling_type=None,
     target_band="mndwi",
     use_viirs=False,
     use_modis=False,
 ):
-    @decorators.carry_metadata
-    def _apply_scaling(img):
-        return img.subtract(min_img).divide(max_img.subtract(min_img)).float()
 
     @decorators.carry_metadata
     def _apply_fusion(img):
@@ -379,23 +168,21 @@ def _fuse_dataset(
     optical = lc8.merge(le7).merge(s2)
 
     if use_viirs:
-        viirs = datasets.Viirs(**ds_kwargs)
+        viirs = datasets.Viirs(**dsa_kwargs)
         optical = optical.merge(viirs)
 
     if use_modis:
         modis = datasets.Modis(**ds_kwargs)
         optical = optical.merge(modis)
 
+    optical = optical.apply_func(geeutils.add_indices,indices=[target_band])
+
     s1 = datasets.Sentinel1(**ds_kwargs)
     s1 = s1.add_fusion_features()
 
-    if scaling_dict is not None:
-        scaling_img = scaling_dict.toImage()
-        min_img = scaling_img.select(".*_min")
-        max_img = scaling_img.select(".*_max")
-        s1 = s1.apply_func(_apply_scaling)
+    s1 = s1.apply_func(ml.standard_image_scaling,scaling_dict,feature_names)
 
-    s1.collection = s1.collection.map(_apply_fusion)
+    s1 = s1.apply_func(_apply_fusion)
 
     fused_ds = optical.merge(s1)
 
@@ -405,10 +192,10 @@ def _fuse_dataset(
         .sort("system:time_start")
     )
 
-    return fused_ds, target_band
+    return fused_ds
 
 
-def export_harmonics(
+def export_surface_water_harmonics(
     region,
     start_time,
     end_time,
@@ -419,6 +206,7 @@ def export_harmonics(
     output_asset_path=None,
     output_bucket=None,
     tile=False,
+    tile_size=1.0
 ):
 
     if tile:
@@ -428,7 +216,7 @@ def export_harmonics(
             .geometry(100)
             .buffer(2500, maxError=100)
         )
-        grid = geeutils.tile_region(region, intersect_geom=land_area, grid_size=1.0)
+        grid = geeutils.tile_region(region, intersect_geom=land_area, grid_size=tile_size)
 
         n = grid.size().getInfo()
         grid_list = grid.toList(n)
@@ -436,25 +224,26 @@ def export_harmonics(
         for i in range(n):
             grid_tile = ee.Feature(grid_list.get(i)).geometry()
             if output_asset_path is not None:
-                output_tile_path = output_asset_path + f"harmonics_t{i}"
+                output_tile_path = output_asset_path + f"harmonics_t{i:05d}"
 
             export_harmonics(
-                grid_tile,
-                start_time,
-                end_time,
-                feature_names,
-                label,
-                fusion_samples,
-                fusion_model_asset,
-                output_tile_path,
-                output_bucket,
+                region=grid_tile,
+                start_time=start_time,
+                end_time=end_time,
+                feature_names=feature_names,
+                label=label,
+                fusion_samples=fusion_samples,
+                fusion_model_asset=fusion_model_asset,
+                output_asset_path=output_tile_path,
+                output_bucket_path=output_bucket,
                 tile=False,
+                tile_size=tile_size
             )
 
     else:
         if fusion_samples is not None:
             fusion_model, scaling_dict = ml.random_forest_ee(
-                25, fusion_samples, feature_names, label, mode="regression"
+                25, fusion_samples, feature_names, label, scaling="standard", mode="regression"
             )
         elif fusion_model_asset is not None:
             raise NotImplementedError()
@@ -463,13 +252,13 @@ def export_harmonics(
                 "Either 'fusion_samples' or 'fusion_model_path' needs to be defined to run fusion process"
             )
 
-        ds, label = _fuse_dataset(
+        ds = _fuse_dataset(
             region,
             start_time,
             end_time,
             fusion_model,
             scaling_dict,
-            target_band="mndwi",
+            target_band=label,
         )
 
         now = datetime.datetime.now()
@@ -490,7 +279,7 @@ def export_harmonics(
         )
 
         harmonic_coefs = timeseries.fit_harmonic_trend(
-            ds, dependent="mndwi", output_err=True
+            ds, dependent=label, output_err=True
         )
         harmonic_coefs = harmonic_coefs.divide(scale_factor).int32().set(metadata)
 
@@ -523,14 +312,15 @@ def export_daily_surface_water(
     look_back=30,
     lag=4,
     output_confidence=False,
-    output_flood = False,
+    output_flood=False,
     fusion_samples=None,
     fusion_model_asset=None,
     output_asset_path=None,
     output_bucket_path=None,
     initial_threshold=0.1,
     tile=False,
-    tile_buffer=100000
+    tile_size=1.0,
+    tile_buffer=100000,
 ):
     def get_weights(i):
         i = ee.Number(i)
@@ -562,9 +352,7 @@ def export_daily_surface_water(
         # uniform sampling of std dev at 95% confidence interval
         long_term_seed = i.add(500)
         short_term_seed = i.add(1000)
-        long_term_random = (
-            ee.Image.random(long_term_seed).multiply(3.92).subtract(1.96)
-        )
+        long_term_random = ee.Image.random(long_term_seed).multiply(3.92).subtract(1.96)
         short_term_random = (
             ee.Image.random(short_term_seed).multiply(3.92).subtract(1.96)
         )
@@ -585,7 +373,7 @@ def export_daily_surface_water(
                 .geometry(100)
                 .buffer(2500, maxError=100)
             )
-            grid = geeutils.tile_region(region, intersect_geom=land_area, grid_size=1.0)
+            grid = geeutils.tile_region(region, intersect_geom=land_area, grid_size=tile_size)
 
             n = grid.size().getInfo()
             grid_list = grid.toList(n)
@@ -602,23 +390,23 @@ def export_daily_surface_water(
 
                 grid_tile = ee.Feature(grid_list.get(i)).geometry()
                 export_daily_surface_water(
-                    grid_tile,
-                    target_date,
-                    harmonic_coefs,
-                    harmonic_collection,
-                    feature_names,
-                    label,
-                    look_back,
-                    lag,
-                    output_confidence,
-                    output_flood,
-                    fusion_samples,
-                    fusion_model_asset,
-                    output_asset_tile,
-                    output_bucket_tile,
-                    initial_threshold,
+                    region=grid_tile,
+                    target_date=target_date,
+                    harmonic_coefs=harmonic_coefs,
+                    harmonic_collection=harmonic_collection,
+                    feature_names=feature_names,
+                    label=label,
+                    look_back=look_back,
+                    lag=lag,
+                    output_confidence=output_confidence,
+                    output_flood=output_flood,
+                    fusion_samples=fusion_samples,
+                    fusion_model_asset=fusion_model_asset,
+                    output_asset_path=output_asset_tile,
+                    output_bucket_path=output_bucket_tile,
+                    initial_threshold=initial_threshold,
                     tile=False,
-                    tile_buffer=tile_buffer
+                    tile_buffer=tile_buffer,
                 )
 
     else:
@@ -628,7 +416,7 @@ def export_daily_surface_water(
 
         if fusion_samples is not None:
             fusion_model, scaling_dict = ml.random_forest_ee(
-                25, fusion_samples, feature_names, label, mode="regression"
+                25, fusion_samples, feature_names, label, scaling="standard", mode="regression"
             )
         elif fusion_model_asset is not None:
             raise NotImplementedError()
@@ -666,7 +454,7 @@ def export_daily_surface_water(
         else:
             harmonic_coefs = harmonic_coefs.select("^(c|t|s).*")
 
-        prod_region = region.buffer(tile_buffer,100)
+        prod_region = region.buffer(tile_buffer, 100)
 
         ds, label = _fuse_dataset(
             prod_region,
@@ -701,34 +489,30 @@ def export_daily_surface_water(
             .reduce("sum")
         )
 
-        fused_pred = (
-            (har_pred.subtract(lin_pred))
-            .convolve(ee.Kernel.gaussian(2.5))
-            .rename("fused_product")
+        fused_pred = hf.filtering.p_median((har_pred.subtract(lin_pred))).rename(
+            "fused_product"
         )
 
-        # water,threshold = thresholding.bmax_otsu(
-        #     fused_pred,
-        #     initial_threshold=0,
-        #     grid_size=0.2,
-        #     region=region,
-        #     invert=True,
-        #     reduction_scale=100,
-        #     return_threshold=True
-        # )
         ci_threshold = thresholding.edge_otsu(
             fused_pred,
             initial_threshold=initial_threshold,
             edge_buffer=300,
             region=prod_region,
             invert=True,
-            reduction_scale=200,
+            scale=150,
             return_threshold=True,
         )
 
-        # ci_threshold = ee.Number(ee.Algorithms.If(ci_threshold.lt(-0.1),-0.1,ci_threshold))
+        permanent_water = (
+            ee.ImageCollection("JRC/GSW1_2/YearlyHistory")
+            .filterDate("1985-01-01",end_time)
+            .limit(5, "system:time_start", False)
+            .map(lambda x: x.select("waterClass").eq(3))
+            .sum()
+            .unmask(0)
+            .gt(0)
+        )
 
-        permanent_water = ee.Image("JRC/GSW1_2/GlobalSurfaceWater").select("occurrence").unmask(0).gt(80)
         water = fused_pred.gt(ci_threshold).Or(permanent_water).rename("water").uint8()
 
         if output_flood:
@@ -760,14 +544,6 @@ def export_daily_surface_water(
                 },
             )
 
-            # ci_threshold = ee.Number(fused_pred.updateMask(water).reduceRegion(
-            #     reducer=ee.Reducer.min(),
-            #     geometry=region,
-            #     scale=100,
-            #     bestEffort=True,
-            #     maxPixels=1e6
-            # ).get('fused_product'))
-
             confidence = (
                 ee.ImageCollection.fromImages(
                     ee.List.sequence(0, 99).map(calc_confidence)
@@ -781,7 +557,6 @@ def export_daily_surface_water(
             out_water = ee.Image.cat([confidence, water,])
         else:
             out_water = water
-        # water.uint8().rename("water"),
 
         fused_pred = fused_pred.multiply(10000).int16()
 
@@ -821,6 +596,7 @@ def export_daily_surface_water(
             bucket = fcomponents[2]
             fpath = fcomponents[3:-1]
 
+            # TODO: remove extension from string formulation
             f_water = "/".join(fpath + [fcomponents[-1] + "_water" + ext])
             f_fusion = "/".join(fpath + [fcomponents[-1] + "_fusion" + ext])
 
@@ -859,7 +635,16 @@ def export_daily_surface_water(
 
     return
 
-def merge_gcp_tiled_results(bucket_path,pattern,region,retries=-1,clean_up=False,cloud_project=None,file_dims=None):
+
+def merge_gcp_tiled_results(
+    bucket_path,
+    pattern,
+    region,
+    retries=-1,
+    clean_up=False,
+    cloud_project=None,
+    file_dims=None,
+):
     land_area = (
         ee.FeatureCollection("USDOS/LSIB_SIMPLE/2017")
         .filterBounds(region)
@@ -872,7 +657,7 @@ def merge_gcp_tiled_results(bucket_path,pattern,region,retries=-1,clean_up=False
 
     fcomponents = bucket_path.split("/")
     bucket = fcomponents[2]
-    fpath = pattern.replace("*","")
+    fpath = pattern.replace("*", "")
 
     files = utils.list_gcs_objs(bucket, pattern=pattern, project=cloud_project)
 
@@ -904,17 +689,20 @@ def merge_gcp_tiled_results(bucket_path,pattern,region,retries=-1,clean_up=False
 
         if clean_up:
             gcsfs.GCSFileSystem.rm(files)
-        
+
     elif retries > 0:
-        time.sleep(60*10)
-        
-        merge_gcp_tiled_results(bucket_path,pattern,region,retries=(retries-1))
+        time.sleep(60 * 10)
+
+        merge_gcp_tiled_results(bucket_path, pattern, region, retries=(retries - 1))
 
     else:
-        raise RuntimeError(f"could not find all expected tiles to merge...found {len(files)} tiles but expected {expected_n}")
+        raise RuntimeError(
+            f"could not find all expected tiles to merge...found {len(files)} tiles but expected {expected_n}"
+        )
 
     return
 
+
 if __name__ == "__main__":
     raise NotImplementedError(
         "Worklow is currently not implemented for CLI use, please check back later"
